#!/usr/bin/python
from __future__ import annotations

from ..module_utils import missing_redirect_target_collection  # pylint: disable=relative-beyond-top-level,unused-import


def main():
    raise Exception('should never get here')


if __name__ == '__main__':
    main()
