using System;

//TypeAccelerator -Name CSRel4 -TypeName TestClass

namespace ansible_collections.my_ns.my_col.plugins.module_utils.sub_pkg.CSRel4
{
    public class TestClass
    {
        public static string Invoke()
        {
            return "CSRel4.Invoke()";
        }
    }
}
